#include <signal.h>

#include <gnome.h>

#include <utillib.h>

#include "dmgmonitor.h"

struct _dmgmonitor
{
	GtkWidget* app;
	GtkWidget* popup;
	GtkMenuShell *projects_menu;
	GSList *projects_group;

	GnomeProgram *gnome_program;
	GnomeClient* gnome_client;

	gboolean connected;
	int fd;

	gint input_tag;

	GList *projects;
	int project;
};

typedef struct _project Project;

struct _project
{
	int id;
	char *name;
	GtkWidget *menuitem;
};

enum {
	TARGET_URI_LIST,
	TARGET_SIMPLE_URL,
	TARGET_TEXT_URL,
	TARGET_HTML_URL
};

static GtkTargetEntry drop_types [] = {
	{ "text/uri-list", 0, TARGET_URI_LIST },
	{ "x-url/http", 0, TARGET_SIMPLE_URL },
	{ "x-url/ftp", 0, TARGET_SIMPLE_URL },
	{ "x-url/*", 0, TARGET_SIMPLE_URL },
	{ "_NETSCAPE_URL", 0, TARGET_SIMPLE_URL },
	{ "text/plain", 0, TARGET_TEXT_URL },
	{ "text/html", 0, TARGET_HTML_URL }
};

static gint n_drop_types = sizeof (drop_types) / sizeof(drop_types[0]);


static gint delete_event_cb(GtkWidget *window, GdkEventAny *e, gpointer data);
void dmgmonitor_get_data(gpointer data, gint source, GdkInputCondition condition);
void dmgmonitor_project_add(dmgMonitor *dmgmonitor, Project *project);
void dmgmonitor_project_change(GtkWidget *widget, gpointer data);
void dmgmonitor_dnd_received (GtkWidget *widget, GdkDragContext *context,
			gint x, gint y, GtkSelectionData *selection_data,
			guint info, guint time, gpointer data);

void dmgmonitor_quit(GtkWidget *caller, gpointer data);
void dmgmonitor_download_new(GtkWidget *caller, gpointer data);

void gdownman_get_data(gpointer data, gint source, GdkInputCondition condition);
void dmgmonitor_ontop_toggle (GtkWidget *caller, gpointer data);

// Menu and toolbar are stored here
#include "dmgmonitor_ui.h"

dmgMonitor *dmgmonitor_new()
{
	dmgMonitor *dmgmonitor = malloc(sizeof(dmgMonitor));

	dmgmonitor->app = 0;
	dmgmonitor->popup = 0;
	dmgmonitor->gnome_program = 0;
	dmgmonitor->gnome_client = 0;
	dmgmonitor->connected = FALSE;
	dmgmonitor->fd = 0;
	dmgmonitor->input_tag = 0;

	dmgmonitor->projects = NULL;
	dmgmonitor->project = 0;

	return dmgmonitor;
}

void dmgmonitor_delete(dmgMonitor *dmgmonitor)
{
	if (dmgmonitor->connected)
		close(dmgmonitor->fd);
}

void dmgmonitor_init(dmgMonitor *dmgmonitor, int argc, char *argv[])
{
	dmgmonitor->gnome_program = gnome_program_init("dmgmonitor", "0.0", LIBGNOMEUI_MODULE,
									argc, argv,
									GNOME_PARAM_POPT_TABLE, NULL, NULL);

	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR)
	{
		perror("can't ignore SIGPIPE");
		exit(1);
	}
}

int dmgmonitor_move(GObject *object, GdkEventButton *event, gpointer data)
{
	gtk_window_begin_move_drag(GTK_WINDOW(object), event->button,
							event->x_root, event->y_root, event->time);
	return TRUE;
}


void dmgmonitor_ui_build(dmgMonitor *dmgmonitor)
{
	GtkWidget *app = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	GtkWidget *image;

	g_signal_connect(G_OBJECT(app), "delete_event",
					G_CALLBACK(delete_event_cb), dmgmonitor);

	gtk_window_set_decorated (GTK_WINDOW(app), FALSE);
	gtk_window_set_default_size(GTK_WINDOW(app), 50, 50);
	gtk_window_set_title(GTK_WINDOW(app), "downman monitor");

	gtk_drag_dest_set (GTK_WIDGET(app), GTK_DEST_DEFAULT_MOTION |
			GTK_DEST_DEFAULT_HIGHLIGHT | GTK_DEST_DEFAULT_DROP,
			drop_types, n_drop_types, GDK_ACTION_COPY);
	gtk_signal_connect (GTK_OBJECT(app), "drag_data_received",
			GTK_SIGNAL_FUNC (dmgmonitor_dnd_received), dmgmonitor);

	dmgmonitor->popup = gnome_popup_menu_new(popup_menu);
	gnome_popup_menu_attach (dmgmonitor->popup, app, dmgmonitor);

	dmgmonitor->projects_menu = 
		GTK_MENU_SHELL(GTK_MENU_ITEM(popup_menu[PROJECTS_SUBMENU].widget)->submenu);

	gtk_widget_add_events (app, GDK_BUTTON_PRESS_MASK);
	
	g_signal_connect(G_OBJECT(app), "button_press_event",
		G_CALLBACK(dmgmonitor_move), NULL);

	image = gtk_image_new_from_file("logo.png");

	gtk_container_add(GTK_CONTAINER(app), image);
	gtk_window_set_icon(GTK_WINDOW(app), gtk_image_get_pixbuf(GTK_IMAGE(image)));

	dmgmonitor->app = app;
}

void dmgmonitor_sm_init(dmgMonitor *dmgmonitor)
{
	dmgmonitor->gnome_client = gnome_master_client();
}

void dmgmonitor_ui_show(dmgMonitor *dmgmonitor)
{
	gtk_widget_show_all(dmgmonitor->app);
}

void dmgmonitor_quit(GtkWidget * caller, gpointer data)
{
	dmgMonitor *dmgmonitor = (dmgMonitor *)data;

	dmgmonitor_delete(dmgmonitor);
	gtk_main_quit();
}

static gint
delete_event_cb(GtkWidget* window, GdkEventAny* e, gpointer data)
{
	dmgmonitor_quit(window, data);

	// Do not destroy it, we did at dmgmonitor_quit()
	return TRUE;
}

void dmgmonitor_project_add(dmgMonitor *dmgmonitor, Project *project)
{
	GtkWidget *menuitem;

	menuitem = gtk_radio_menu_item_new_with_label(
							dmgmonitor->projects_group, project->name);

	g_object_set_data(G_OBJECT(menuitem), "project", project);
	dmgmonitor->projects_group = gtk_radio_menu_item_group(
									GTK_RADIO_MENU_ITEM (menuitem));
	gtk_menu_shell_append(GTK_MENU_SHELL(dmgmonitor->projects_menu), menuitem);
	gtk_widget_show(menuitem);

	g_signal_connect(G_OBJECT(menuitem), "toggled",
					GTK_SIGNAL_FUNC(dmgmonitor_project_change), dmgmonitor);

	project->menuitem = menuitem;
}

void dmgmonitor_project_change(GtkWidget *widget, gpointer data)
{
	dmgMonitor *dmgmonitor = (dmgMonitor *) data;
	Project *project = (Project *) g_object_get_data(G_OBJECT(widget), "project");

	if (GTK_CHECK_MENU_ITEM(widget)->active)
		dmgmonitor->project = project->id;
}


void dmgmonitor_connect_local(dmgMonitor *dmgmonitor)
{
	dmgmonitor->fd = dm_client_local_connect("downman_gmonitor");

//FIXME: Try autostart before reporting error
	if (dmgmonitor->fd < 0)
	{
		GtkWidget *dialog = gtk_message_dialog_new (GTK_WINDOW(dmgmonitor->app),
								GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR,
								GTK_BUTTONS_OK, "%s", "There was an error while connecting to the daemon. Please, check that downmand is running.");

		gtk_dialog_run(GTK_DIALOG(dialog));
		gtk_widget_destroy(dialog);

//FIXME: start a gtk_timeout to try to connect to the local daemon (every 1s?).
		gtk_widget_set_sensitive(GTK_WIDGET(GTK_MENU_ITEM(
						popup_menu[NEW_DOWNLOAD].widget)), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(GTK_MENU_ITEM(
						popup_menu[PROJECTS_SUBMENU].widget)), FALSE);
		return;
	}

	dmgmonitor->connected = TRUE;

	dmgmonitor->input_tag = gdk_input_add(dmgmonitor->fd, GDK_INPUT_READ,
										dmgmonitor_get_data, dmgmonitor);
	dm_command_events(dmgmonitor->fd, 1, NULL);
	dm_command_batchstatus(dmgmonitor->fd, 1, NULL);
}

void dmgmonitor_download_new(GtkWidget *caller, gpointer data)
{
	dmgMonitor *dmgmonitor = (dmgMonitor *)data;
	GtkWidget *dialog, *label, *entry;
	const gchar *url;

	dialog = gtk_dialog_new_with_buttons ("New Download",
					GTK_WINDOW(dmgmonitor->app), 0,
					GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
					GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);

	label = gtk_label_new("<big>Enter a new <b>URL</b></big>");
	gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
	entry = gtk_entry_new();

	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);

	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), entry);
	gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
	gtk_widget_grab_focus(entry);

	gtk_widget_show_all (dialog);

	switch(gtk_dialog_run(GTK_DIALOG(dialog)))
	{
		case GTK_RESPONSE_ACCEPT:
			url = gtk_entry_get_text(GTK_ENTRY(entry));
			if (strlen(url) > 0)
                        {
                            GString *item;
                            item = dm_command_get_new (dmgmonitor->project,
                                                       url);
                            dm_command_get (item, dmgmonitor->fd);
                        }
			break;

		default:
			break;
	}

	gtk_widget_destroy(dialog);
}

void
dmgmonitor_ontop_toggle (GtkWidget *caller, gpointer data)
{
    dmgMonitor *dmgmonitor = (dmgMonitor *)data;

//FIXME: This is already in gtk 2.3
#if 0
    if (GTK_CHECK_MENU_ITEM (caller)->active)
    {
        gtk_window_set_keep_above (GTK_WINDOW (dmgmonitor->app), TRUE);
    }
    else
    {
        gtk_window_set_keep_above (GTK_WINDOW (dmgmonitor->app), FALSE);
    }
#endif
}

void dmgmonitor_get_data(gpointer data, gint source, GdkInputCondition condition)
{
	dmgMonitor *dmgmonitor = (dmgMonitor *) data;
	unsigned int project_id;
	char buffer[4096];
	char s[2048];
	long int i,j;
	double k,l;
	double percent;
	int row = 0;
	int readed;
	char ch;

	do
	{
		readed = read(dmgmonitor->fd, &ch, 1);
		if (readed < 1)
			return;
		buffer[row++] = ch;
	} while (ch != '\n');

	buffer[row - 1] = '\0'; 

	printf("GOT: %s\n", buffer);

	if (sscanf(buffer, "%li", &i) == 0)
		return;

	switch(i)
	{
		case DM_COMMAND_BATCH_STATUS:
			if (sscanf(buffer + 4, "%u %i %i %lf %lf",
					&project_id, &i, &j, &k, &l) < 5)
				return;

			if (l > 0)
				percent = (k/l)*100;
			else
				percent = 0;

			if (dmgmonitor->project == project_id)
			{
				snprintf(buffer, 2000,"%i/%i\n%.1f%%", i, i + j, percent);
//FIXME: composite on top of the pixmap
				//gtk_label_set_text(GTK_LABEL(dmgmonitor->label), buffer);
			}
			break;

		case DM_COMMAND_PROJECT_NEW:
			if (sscanf(buffer + 4, "%i %s", &project_id, s) < 2)
				return;
			else
			{
				Project *project = malloc(sizeof(Project));
				project->id = project_id;
				project->name = strdup(s);

                dmgmonitor->projects = g_list_append (dmgmonitor->projects,
                                                      project);
                dmgmonitor_project_add (dmgmonitor, project);
				if (dmgmonitor->project == 0)
					dmgmonitor->project = project_id;
			}
			break;

		default:
			printf("Not handled response!!!\n");
			break;
	}
}

void dmgmonitor_dnd_received (GtkWidget *widget, GdkDragContext *context,
			gint x, gint y, GtkSelectionData *selection_data,
			guint info, guint time, gpointer data)
{
    GString *item;
	char *buf;
	dmgMonitor *dmgmonitor = (dmgMonitor *) data;

	buf = strchr(selection_data->data, '\n');
	if (buf)
		buf[0] = '\0';

    item = dm_command_get_new (dmgmonitor->project, selection_data->data);
    dm_command_get (item, dmgmonitor->fd);

	printf("URL: %s\n", selection_data->data);
}
