#include <stdio.h>
#include <string.h>

#include <glib.h>

#include <utillib.h>

#include "downmand-marshal.h"
#include "main.h"
#include "debug.h"
#include "client.h"

#define MAX_COMMAND_LENGTH 10000

GType           client_get_type           ();
static void     client_class_init         (ClientClass *klass);
static void     client_init               (Client *client);
static void     client_finalize           (GObject* object);

void        client_mutex_lock       (Client *client);
void        client_mutex_unlock     (Client *client);
gboolean    client_input            (GIOChannel *source, GIOCondition condition,
                                     gpointer data);


enum {
    CLIENT_COMMAND_RECEIVED,
    CLIENT_DISCONNECT,
    LAST_SIGNAL
};

static gint client_signals[LAST_SIGNAL] = { 0 };

static gpointer parent_class = NULL;


GType client_get_type ()
{
    static GType client_type = 0;

    if (!client_type)
    {
        static const GTypeInfo client_info =
        {
            sizeof (ClientClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) client_class_init,
            NULL,           /* class_finalize */
            NULL,           /* class_data */
            sizeof (Client),
            0,              /* n_preallocs */
            (GInstanceInitFunc) client_init
        };

        client_type = g_type_register_static (G_TYPE_OBJECT,
                                               "Client",
                                               &client_info,
                                               0);
    }

    return client_type;
}

static void client_class_init (ClientClass *klass)
{
    GObjectClass *object_class;

    object_class = (GObjectClass*) klass;

    parent_class = g_type_class_peek_parent (klass);

    client_signals[CLIENT_COMMAND_RECEIVED] = g_signal_new ("command_received",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ClientClass, client_command_received),
            NULL, NULL,
            downmand_marshal_VOID__POINTER,
            G_TYPE_NONE,
            1, G_TYPE_POINTER);

    client_signals[CLIENT_DISCONNECT] = g_signal_new ("disconnect",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (ClientClass, client_disconnect),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    klass->client_command_received = NULL;
    klass->client_disconnect = NULL;

    object_class->finalize  = client_finalize;
}

static void client_init (Client *client)
{
    client->id = 0;
    client->name = g_string_new (NULL);

    client->buffer = g_string_new (NULL);
    client->buffer_in = g_string_new (NULL);
    client->buffer_out = g_string_new (NULL);
    client->buffer_read = g_string_new (NULL);

    g_static_mutex_init (&client->client_mutex);

    client->output_busy = 0;
    client->output_messages = g_queue_new ();
    client->batch_status = 0;
    client->batch_progress = 0;
    client->events = 0;
    client->events_projects = 0;
}

static void
client_finalize (GObject* object)
{
    Client *client = CLIENT (object);

    g_string_free(client->name, TRUE);
    g_io_channel_close(client->socket);

    g_string_free(client->buffer, TRUE);
    g_string_free(client->buffer_in, TRUE);
    g_string_free(client->buffer_out, TRUE);
    g_string_free(client->buffer_read, TRUE);

    if (G_OBJECT_CLASS (parent_class)->finalize)
        (* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

Client *client_new (gint fd)
{
    Client *client = CLIENT (g_object_new (client_get_type (), NULL));

    client->socket = g_io_channel_unix_new (fd);
    g_io_channel_set_flags (client->socket, G_IO_FLAG_NONBLOCK, NULL);
    client->connected = 1;

    client->input_tag = g_io_add_watch(client->socket,
                    G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
                    client_input, client);

	return client;
}

gboolean
client_input (GIOChannel *source, GIOCondition condition, gpointer data)
{
    Client *client = CLIENT (data);
    GIOStatus status;
    GError *err = NULL;
    gsize terminator_pos;
    GString *buffer = client->buffer_in;
    GString *read_buffer = client->buffer_read;

    if ((condition & G_IO_ERR) || (condition & G_IO_PRI) ||
        (condition & G_IO_NVAL) || (condition & G_IO_OUT))
    {
        g_source_remove (client->output_tag);
        g_signal_emit (G_OBJECT (client), client_signals[CLIENT_DISCONNECT], 0);
        return FALSE;
    }
    else if (condition & G_IO_IN)
    {
        status = g_io_channel_read_line_string(client->socket, read_buffer,
                                               &terminator_pos, &err);
        g_string_append (buffer, read_buffer->str);

        /* Disconnect client on too large commands */
        if (buffer->len > MAX_COMMAND_LENGTH)
            status = G_IO_STATUS_EOF;

        if (status == G_IO_STATUS_EOF)
        {
            if (err != NULL)
            {
                /* Report error to user, and free error */
                g_printerr ("EOF reading data: %s\n", err->message);
                g_error_free (err);
            }
            g_signal_emit (G_OBJECT (client), client_signals[CLIENT_DISCONNECT],
                           0);
            return FALSE;
        }
        else if (status == G_IO_STATUS_ERROR)
        {
            if (err != NULL)
            {
                /* Report error to user, and free error */
                g_printerr ("Error reading data: %s\n", err->message);
                g_error_free (err);
            }
            g_signal_emit (G_OBJECT (client), client_signals[CLIENT_DISCONNECT],
                           0);
            return FALSE;
        }
        else if (status == G_IO_STATUS_AGAIN)
        {
            /* More data is needed, return TRUE */
            return TRUE;
        }
        else
        {
            /* g_print ("status = %i\n", status); */
            buffer->str[terminator_pos] = '\0';
            g_signal_emit (G_OBJECT (client),
                           client_signals[CLIENT_COMMAND_RECEIVED],
                           0,
                           buffer->str);
            g_string_truncate (buffer, 0);
        }

        return TRUE;
    }
    else if (condition & G_IO_HUP)
    {
        g_signal_emit (G_OBJECT (client), client_signals[CLIENT_DISCONNECT], 0);
        return FALSE;
    }

    /* We shouldn't get here */
    {
        gchar *str;
        status = g_io_channel_read_to_end(client->socket, &str,
                                            &terminator_pos, &err);
        if (err != NULL)
        {
            /* Report error to user, and free error */
            g_printerr ("Error reading data: %s\n", err->message);
            g_error_free (err);
        }

        g_print("DATA: %s\n", str);
        g_free(str);
    }

    g_signal_emit (G_OBJECT (client), client_signals[CLIENT_DISCONNECT], 0);
    g_return_val_if_reached(FALSE);
}

gboolean
client_output (GIOChannel *source, GIOCondition condition, gpointer data)
{
    Client *client = CLIENT (data);
    GIOStatus status;
    GError *err = NULL;
    gchar *message;
    gsize bytes_written;

    if (g_queue_is_empty (client->output_messages))
        return FALSE;

    do {
        message = g_queue_pop_head (client->output_messages);
        status = g_io_channel_write_chars (client->socket, message,
                                           strlen (message),
                                           &bytes_written,
                                           &err);
    } while ((status == G_IO_STATUS_NORMAL)
             && (!g_queue_is_empty (client->output_messages)));

    if (g_queue_is_empty (client->output_messages))
    {
        return FALSE;
    }
    else
    {
        return TRUE;
    }
}

int client_parse_command(Client *client, char *buffer)
{
/*
		// Misc command
		case 'M':
			buffer++;
			if (strncmp(buffer, "DISCONNECT", 10) == 0)
			{
				// If we receive a disconnect command return 1 so
				// that the calling code knows how to delete the client.
				printf("%s Disconnected\n", client->name->str);
//FIXME
//				return 666;
			}
			else if (strncmp(buffer, "SHUTDOWN", 8) == 0)
			{
				printf("SHUTDOWN requested by %s.\n", client->name->str);
//FIXME
				downmand_shutdown_set(downmand, 1);			
			}
	}
*/
    return 0;
}

void
client_send (Client *client, gchar *buffer)
{
	GIOStatus status;
	GError *err = NULL;
	gsize bytes_written;

    if (g_queue_is_empty (client->output_messages))
    {
	    status = g_io_channel_write_chars (client->socket, buffer,
                                           strlen(buffer),
                                           &bytes_written,
                                           &err);

        if (status != G_IO_STATUS_NORMAL)
        {
            if (status == G_IO_STATUS_AGAIN)
            {
                g_queue_push_tail (client->output_messages, g_strdup (buffer));
                client->output_tag = g_io_add_watch (client->socket, G_IO_OUT,
                                                     client_output,
                                                     client);
                return;
            }
        }

    	if (err != NULL)
	    {
		    /* Report error to user, and free error */
    		g_debug ("Error writing to client: %s\n", err->message);
            g_error_free (err);
        }

    	status = g_io_channel_flush(client->socket, &err);
    	if (err != NULL)
	    {
		    /* Report error to user, and free error */
    		g_debug ("Error flushing data to the client: %s\n", err->message);
    		g_error_free (err);
    	}
    }
    else
    {
        g_queue_push_tail (client->output_messages, g_strdup (buffer));
    }
}

void
client_mutex_lock (Client *client)
{
    g_static_mutex_lock (&client->client_mutex);
}

void
client_mutex_unlock (Client *client)
{
    g_static_mutex_unlock (&client->client_mutex);
}

guint
client_events (Client *client)
{
    return client->events;
}

guint
client_events_projects (Client *client)
{
    return client->events_projects;
}

guint
client_batch_status (Client *client)
{
    return client->batch_status;
}

guint
client_batch_progress (Client *client)
{
    return client->batch_progress;
}
