#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <signal.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include <glib.h>
#include <libgnomevfs/gnome-vfs-init.h>

#include <utillib.h>

#include "main.h"
#include "debug.h"
#include "downmand.h"
#include "project.h"
#include "transfer.h"
#include "client.h"

GType           downmand_get_type           ();
static void     downmand_class_init         (DownmandClass *klass);
static void     downmand_init               (Downmand *downmand);
gboolean        downmand_projects_scheduler (gpointer data);
gboolean        downmand_projects_save      (gpointer data);
gboolean        downmand_new_client         (GIOChannel *source,
                                             GIOCondition condition,
                                             gpointer data);
gint            create_dir                  (GString *buffer);

void    downmand_command_get_project_and_transfer
                                        (Downmand *downmand, gchar *command,
                                         Project **project,
                                         guint *transfer_id,
                                         GError **error);
void    downmand_command_get_project_transfer_and_setting
                                        (Downmand *downmand,
                                         gchar *command,
                                         Project **project,
                                         guint *transfer_id,
                                         guint *setting,
                                         GError **error);
void    downmand_command_project_new    (Downmand *downmand, gchar *name,
                                         GError **error);
void    downmand_command_project_delete (Downmand *downmand, Project *project,
                                         GError **error);
void    downmand_command_get_project    (Downmand *downmand, gchar *command,
                                         Project **project,
                                         GError **error);
void    downmand_command_get_project_and_setting
                                        (Downmand *downmand, gchar *command,
                                         Project **project,
                                         guint *setting,
                                         GError **error);
void    downmand_command_get_setting    (gchar *command, guint *setting,
                                         GError **error);


void    downmand_project_autotransfer_send  (Project *project, gpointer data);
void    downmand_project_maxtransfers_send  (Project *project, gpointer data);
void    downmand_transfer_new       (Project *project, Transfer *transfer,
                                     gpointer data);
void    downmand_transfer_stop      (Project *project, Transfer *transfer,
                                     gpointer data);
void    downmand_transfer_done      (Project *project, Transfer *transfer,
                                     gpointer data);
void    downmand_transfer_delete    (Project *project, Transfer *transfer,
                                     gpointer data);
void    downmand_transfer_update    (Project *project, Transfer *transfer,
                                     gint setting,
                                     gpointer data);

void    downmand_servers_new        (Servers *servers, Server *server,
                                     gpointer data);
void    downmand_servers_changed    (Servers *servers, Server *server,
                                     gint setting,
                                     gpointer data);

void    downmand_clients_servers_send   (Downmand *downmand, Client *client);
void    downmand_send_to_clients            (Downmand *downmand,
                                             gchar *command);
void    downmand_project_send_to_clients    (Downmand *downmand,
                                             gchar *command);

void    downmand_project_new            (Downmand *downmand, Project *project,
                                         GError **error);

void    downmand_clients_project_new    (Downmand *downmand, Project *project);
void    downmand_clients_project_delete (Downmand *downmand, Project *project);
void    downmand_clients_batch_send     (Downmand *downmand,
                                         Client *client);
void    downmand_clients_projects_send  (Downmand *downmand,
                                         Client *client);
void    downmand_client_delete          (GObject *caller, gpointer data);

void    downmand_shutdown                   (Downmand *downmand);
void    downmand_shutdown_transfers_stop    (Downmand *downmand);


static gpointer parent_class = NULL;

GType
downmand_get_type ()
{
    static GType downmand_type = 0;

    if (!downmand_type)
    {
        static const GTypeInfo downmand_info =
        {
            sizeof (DownmandClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) downmand_class_init,
            NULL,           /* class_finalize */
            NULL,           /* class_data */
            sizeof (Downmand),
            0,              /* n_preallocs */
            (GInstanceInitFunc) downmand_init
        };

        downmand_type = g_type_register_static (G_TYPE_OBJECT,
                                                "Downmand",
                                                &downmand_info,
                                                0);
    }

    return downmand_type;
}

static void
downmand_class_init (DownmandClass *klass)
{
    parent_class = g_type_class_peek_parent (klass);
}

static void
downmand_init (Downmand *downmand)
{
    downmand->clients = g_hash_table_new (NULL, NULL);
    downmand->projects = g_hash_table_new (NULL, NULL);
    downmand->local_socket = 0;
    downmand->g_local_socket = 0;
    downmand->count = 0;
    downmand->project_id = 1;
    downmand->shutdown = 0;
    downmand->servers = servers_new ();
    downmand->clients_count = 0;
    downmand->buffer = g_string_new (NULL);
}

Downmand *
downmand_new()
{
    Downmand *downmand = g_object_new (downmand_get_type (), NULL);

    gnome_vfs_init();

    return downmand;
}

/* check that needed directories exist */
gint
downmand_workspace_init (GError **error)
{
	GString *buffer = g_string_new (NULL);
	int errorno = 0;

	g_string_printf (buffer, "%s/.downman", g_get_home_dir ());
	errorno = create_dir(buffer);
	if (errorno)
	{
		g_set_error (error, DM_DOWNMAND_ERROR, 1,
			"Can't create directory %s: %s.",
			buffer->str, g_strerror (errorno));
		return 1;
	}

	g_string_printf (buffer, "%s/.downman/projects", g_get_home_dir ());
	errorno = create_dir(buffer);
	if (errorno)
	{
		g_set_error (error, DM_DOWNMAND_ERROR, 1,
			"Can't create directory %s: %s.",
			buffer->str, g_strerror (errorno));
		return 1;
	}

	return 0;
}

gint
create_dir (GString *buffer)
{
	int errorno = 0;

	if (access (buffer->str, R_OK) == -1)
	{
		int old_uid;

		old_uid = geteuid();
		seteuid (getuid());
		if (mkdir (buffer->str, 0755) == -1)
		{
			errorno = errno;
		}
		seteuid (old_uid);
	}

	return errorno;
}

guint
downmand_project_id(Downmand *downmand)
{
	return downmand->project_id++;
}

/* returns 0: Ok, 1: error */
gint
downmand_socket_local_create (Downmand *downmand, GError **error)
{
	int local_socket_fd;
	struct sockaddr_un addr;
	size_t size;
	int flags, old_uid;
	GString *buffer = downmand->buffer;

	// drop root privileges before creating the socket
	old_uid = geteuid();
	seteuid(getuid());

	g_string_printf(buffer, "/tmp/downmand.%s", g_get_user_name ());

	// configure the socket
	addr.sun_family = AF_FILE;
	strcpy(addr.sun_path, buffer->str);
	size = sizeof(addr.sun_family) + strlen(addr.sun_path) + 1;

	if (access(buffer->str, R_OK) != -1)
	{
		int sock_fd;

		// If the socket already exists, try to connect...
		sock_fd = socket(PF_UNIX, SOCK_STREAM, 0);
		if (sock_fd < 0)
		{
			g_set_error (error, DM_DOWNMAND_ERROR, 1,
				"Failed to create a socket: %s.",
				g_strerror (errno));
			return 1;
		}
		if (connect(sock_fd, (struct sockaddr *)&addr, size) == 0)
		{
			// If we connect, tell our name and exit (already running)
			write(sock_fd, "CNAME downmand\n", 14);
			close(sock_fd);
			g_set_error (error, DM_DOWNMAND_ERROR, 2,
				"downmand already running! No need to run again.");
			return 2;
		}

		// If we can't, the daemon is dead.
		if(unlink(buffer->str) == -1)
		{
			g_set_error (error, DM_DOWNMAND_ERROR, 3,
				"Can't delete client local socket: %s.",
				g_strerror (errno));
			return 3;
		}
	}

	// if we get here, we must create the default socket       
	local_socket_fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (local_socket_fd < 0)
	{
		seteuid(old_uid);
		g_set_error (error, DM_DOWNMAND_ERROR, 4,
			"Error creating client local socket: %s.",
			g_strerror (errno));
		return 4;
	}

	flags = fcntl(local_socket_fd, F_GETFL, 0);
	if (flags == -1)
	{
		seteuid(old_uid);
		g_set_error (error, DM_DOWNMAND_ERROR, 5,
			"Can't read socket flags: %s.",
			g_strerror (errno));
		return 5;
	}
	flags |= O_NONBLOCK;
	if (fcntl(local_socket_fd, F_SETFL, flags) < 0)
	{
		seteuid(old_uid);
		g_set_error (error, DM_DOWNMAND_ERROR, 6,
			"Can't set socket flags: %s.",
			g_strerror (errno));
		return 6;
	}

	if (bind(local_socket_fd, (struct sockaddr *)&addr, size) < 0)
	{
		seteuid(old_uid);
		g_set_error (error, DM_DOWNMAND_ERROR, 7,
			"Can't bind client local socket: %s.",
			g_strerror (errno));
		return 7;
	}

	// make it only read-write to the efective user
	if (chmod(buffer->str, S_IRUSR | S_IWUSR) < 0)
	{
		seteuid(old_uid);
		g_set_error (error, DM_DOWNMAND_ERROR, 8,
			"Can't set client local socket to 400 mode: %s.",
			g_strerror (errno));
		return 8;
	}

	if (listen(local_socket_fd, 1) < 0)
	{
		seteuid(old_uid);
		g_set_error (error, DM_DOWNMAND_ERROR, 9,
			"Can't listen on client local socket: %s.",
			g_strerror (errno));
		return 9;
	}

    seteuid (old_uid);
    downmand->local_socket = local_socket_fd;
    downmand->g_local_socket = g_io_channel_unix_new (local_socket_fd);

	return 0;	
}

/* Commands *******************************************************************/

void
downmand_command_handle (GObject *caller, gchar *command, gpointer data)
{
    Downmand *downmand = DOWNMAND (data);
    Project *project;
    Client *client = CLIENT (caller);
    GError *err = NULL;
    guint command_id;
    guint transfer_id;
    guint setting;

	g_return_if_fail(command != NULL);

//    g_print("GOT: %s\n", command);

    if (sscanf(command, "%u ", &command_id) == 0)
    {
//FIXME: send "BAD COMMAND" error to client
        return;
    }

    switch (command_id)
    {
        /* Transfer commands */
        case DM_COMMAND_ITEM_NEW:
            downmand_command_get_project (downmand, command + 4, &project,
                                          &err);
            if (err == NULL)
                project_command_transfer_new (project,
                                              strchr (command + 5, ' ') + 1,
                                              &err);
            break;

        case DM_COMMAND_ITEM_START:
            downmand_command_get_project_and_transfer (downmand, command + 4,
                                                       &project,
                                                       &transfer_id,
                                                       &err);
            if (err == NULL)
                project_command_transfer_start (project, transfer_id, &err);
            break;

        case DM_COMMAND_ITEM_STOP:
            downmand_command_get_project_and_transfer (downmand, command + 4,
                                                       &project,
                                                       &transfer_id,
                                                       &err);
            if (err == NULL)
                project_command_transfer_stop (project, transfer_id, &err);
            break;

        case DM_COMMAND_ITEM_QUEUE:
            downmand_command_get_project_and_transfer (downmand, command + 4,
                                                       &project,
                                                       &transfer_id,
                                                       &err);
            if (err == NULL)
                project_command_transfer_queue (project, transfer_id, &err);
            break;

        case DM_COMMAND_ITEM_DELETE:
            downmand_command_get_project_and_transfer (downmand, command + 4,
                                                       &project,
                                                       &transfer_id,
                                                       &err);
            if (err == NULL)
                project_command_transfer_delete (project, transfer_id, &err);
            break;

        case DM_COMMAND_ITEM_SPEED:
            downmand_command_get_project_transfer_and_setting (downmand,
                                                       command + 4,
                                                       &project,
                                                       &transfer_id,
                                                       &setting,
                                                       &err);
            if (err == NULL)
                project_command_transfer_speed (project, transfer_id, setting,
                                                &err);
            break;

        /* Servers commands */
        case DM_COMMAND_SERVER_EVENTS:
            downmand_command_get_setting (command + 4, &setting, &err);
            if (err == NULL)
            {
                if (setting <= 0)
                    client->events_servers = 0;
                else
                {
                    client->events_servers = setting;
                    downmand_clients_servers_send (downmand, client);
                }
            }
            break;

        case DM_COMMAND_SERVER_MAX_CONNECTIONS:
            servers_max_connections (downmand->servers, command + 4);
            break;

        /* Project commands */
        case DM_COMMAND_PROJECT_NEW:
            downmand_command_project_new (downmand, command + 4, &err);
            break;

        case DM_COMMAND_PROJECT_DELETE:
            downmand_command_get_project (downmand, command + 4, &project,
                                          &err);
            if (err == NULL)
                downmand_command_project_delete (downmand, project, &err);
            break;

        case DM_COMMAND_PROJECT_AUTO:
            downmand_command_get_project_and_setting (downmand, command + 4,
                                                      &project,
                                                      &setting,
                                                      &err);
            if (err == NULL)
                project_auto_set (project, setting, &err);
            break;

        case DM_COMMAND_PROJECT_MAXTRANSFERS:
            downmand_command_get_project_and_setting (downmand, command + 4,
                                                      &project,
                                                      &setting,
                                                      &err);
            if (err == NULL)
                project_maxtransfers_set (project, setting, &err);
            break;

        /* Client configuration commands */
        case DM_COMMAND_CONFIG_NAME:
            client->name = g_string_assign (client->name, command + 5);
			client_send(client, "NAME OK\n");
            break;

        case DM_COMMAND_CONFIG_EVENTS:
            downmand_command_get_setting (command + 4, &setting, &err);
            if (err == NULL)
            {
                if (setting <= 0)
                    client->events = 0;
                else
                {
                    client->events = setting;
                    downmand_clients_batch_send(downmand, client);
                }
            }
            break;

        case DM_COMMAND_CONFIG_PROJECT_EVENTS:
            downmand_command_get_setting (command + 4, &setting, &err);
            if (err == NULL)
            {
                if (setting <= 0)
                    client->events_projects = 0;
                else
                {
                    client->events_projects = setting;
                    downmand_clients_projects_send(downmand, client);
                }
            }
            break;

        case DM_COMMAND_CONFIG_BATCHSTATUS:
            downmand_command_get_setting (command + 4, &setting, &err);
            if (err == NULL)
            {
                if (setting <= 0)
                    client->batch_status = 0;
                else
                {
                    client->batch_status = setting;
                }
            }
            break;

        case DM_COMMAND_CONFIG_BATCHPROGRESS:
            downmand_command_get_setting (command + 4, &setting, &err);
            if (err == NULL)
            {
                if (setting <= 0)
                    client->batch_progress = 0;
                else
                {
                    client->batch_progress = setting;
                }
            }
            break;

        default:
            g_set_error (&err, DM_DOWNMAND_ERROR, 1,
                "Unsupported command: %s", command);
            g_print (err->message);
            break;
    }

    if (err != NULL)
    {
        GString *buffer = downmand->buffer;

        g_string_printf (buffer, "%u %s\n", DM_COMMAND_ERROR, err->message);
        client_send (client, buffer->str);
    }
}

/* Extracts the project and the transfer id from the command */
void
downmand_command_get_project_and_transfer (Downmand *downmand, gchar *command,
                                           Project **project,
                                           guint *transfer_id,
                                           GError **error)
{
	guint project_id;

    if (sscanf(command, "%u %u", &project_id, transfer_id) > 1)
    {
        if ((project_id > 0) && (*transfer_id > 0))
        {
            *project = g_hash_table_lookup (downmand->projects,
                                            GINT_TO_POINTER (project_id));

            if (*project == NULL)
            {
                g_set_error (error, DM_DOWNMAND_ERROR, 1,
                    "Project %i does not exist", project_id);
            }
        }
        else
        {
            g_set_error (error, DM_DOWNMAND_ERROR, 1,
                "Command does not contain a valid project or transfer id");
        }
    }
    else
    {
        g_set_error (error, DM_DOWNMAND_ERROR, 1,
            "Command does not contain a project or transfer id");
    }
}

/* Extracts the project, the transfer id and the setting from the command */
void
downmand_command_get_project_transfer_and_setting (Downmand *downmand, gchar *command,
    Project **project,
    guint *transfer_id,
    guint *setting,
    GError **error)
{
	guint project_id;

    if (sscanf(command, "%u %u %u", &project_id, transfer_id, setting) > 1)
    {
        if ((project_id > 0) && (*transfer_id > 0))
        {
            *project = g_hash_table_lookup (downmand->projects,
                                            GINT_TO_POINTER (project_id));

            if (*project == NULL)
            {
                g_set_error (error, DM_DOWNMAND_ERROR, 1,
                    "Project %i does not exist", project_id);
            }
        }
        else
        {
            g_set_error (error, DM_DOWNMAND_ERROR, 1,
                "Command does not contain a valid project or transfer id");
        }
    }
    else
    {
        g_set_error (error, DM_DOWNMAND_ERROR, 1,
            "Command does not contain a project or transfer id");
    }
}

gint
mkdir_with_parents (const char *pathname, mode_t mode)
{
    GString *buffer = g_string_new (NULL);
    gchar **paths;
    gchar *path;
    gint i = 0;

    paths = g_strsplit (pathname, "/", 0);

    for (path = paths[i]; path != NULL; path = paths[++i])
    {
        g_string_append_printf (buffer, "/%s", path);
        if (!g_file_test (buffer->str, G_FILE_TEST_EXISTS))
        {
            mkdir (buffer->str, mode);
        }
    }

    return 0;
}

void
downmand_command_project_new (Downmand *downmand, gchar *command,
    GError **error)
{
    Project *project;
    GString *buffer = downmand->buffer;
    GError *err = NULL;
    gchar **commands;
    guint command_id = 0, value = 0;
    gint i = 0;

    project = project_new (downmand_project_id (downmand), &err);

    if (err != NULL)
    {
        g_propagate_error (error, err);
        return;
    }

    commands = g_strsplit (command, "|", 0);

    for (command = commands[i]; command != NULL; command = commands[++i])
    {
        if (sscanf(command, "%u ", &command_id) == 0)
        {
//FIXME: project_log_add
//            project_log_add (project, "* Bad command id: %s\n", command);
            continue;
        }

        switch (command_id)
        {
            /* Transfer commands */
            case DM_COMMAND_PROJECT_NAME:
                project_name_set (project, command + 4);
                break;

            case DM_COMMAND_PROJECT_PATH:
//FIXME: project_log_add
//                project_log_add (project, "* Set PATH: %s\n", command + 4);
                project_default_path_set (project, command + 4);
                break;

            case DM_COMMAND_PROJECT_AUTO:
                if (sscanf (command + 4, "%u", &value) > 0)
                {
//FIXME: project_log_add
//                    project_log_add (project,
//                                  "* Set FLAGS: AUTO: %u.\n", value);
                    project_auto_set (project, value, &err);
                }
                break;

            case DM_COMMAND_PROJECT_SPEED:
                if (sscanf (command + 4, "%u", &value) > 0)
                {
//FIXME: project_log_add
//                    project_log_add (project,
//                                  "* Set FLAGS: Speed limit on: %u.\n", value);
                    project_speedlimit_set (project, value);
                }
                break;

            default:
//FIXME: project_log_add
//                project_log_add (project, "COMMAND NOT UNDERSTOOD: %s\n", command);
                break;
        }
    }

    g_strfreev(commands);

    if (project_name_get (project) == 0)
    {
//FIXME: Report the error
        return;
    }

    if (project_default_path_get (project) == 0)
    {
//FIXME: Report the error
        return;
    }

    /* check directories */
    g_string_printf (buffer, "%s/.downman/projects/%s/", g_get_home_dir (),
                     project_name_get (project));
    if (g_file_test (buffer->str, G_FILE_TEST_EXISTS))
    {
        g_set_error (error, DM_DOWNMAND_ERROR, 1,
                     "The project %s already exists",
                     project_name_get (project));
        return;
    }

    downmand_project_new (downmand, project, &err);

    if (err != NULL)
    {
        g_propagate_error (error, err);
        return;
    }

    /* notice clients */
    downmand_clients_project_new (downmand, project);
    downmand_project_autotransfer_send (project, downmand);
    downmand_project_maxtransfers_send (project, downmand);
}

void
downmand_project_new (Downmand *downmand, Project *project, GError **error)
{
    GString *buffer = downmand->buffer;

    /* check directories */
    g_string_printf (buffer, "%s/.downman/projects/%s/", g_get_home_dir (),
                     project_name_get (project));
    if (!g_file_test (buffer->str, G_FILE_TEST_EXISTS))
    {
        mkdir (buffer->str, S_IRWXU);
    }

    project_path_set (project, buffer->str);

    g_string_printf (buffer, "%s", project_default_path_get (project));
    if (g_file_test (buffer->str, G_FILE_TEST_EXISTS))
    {
//FIXME: Check if writable
    }
    else
    {
        mkdir_with_parents (buffer->str, S_IRWXU);
    }

    /* Settings */
    project_preferences_load (project);
    project_session_load (project);

    g_hash_table_insert (downmand->projects,
                         GINT_TO_POINTER (project_id (project)),
                         project);

    g_signal_connect (project, "project_autotransfer_changed",
                      G_CALLBACK (downmand_project_autotransfer_send),
                      downmand);

    g_signal_connect (project, "project_maxtransfers_changed",
                      G_CALLBACK (downmand_project_maxtransfers_send),
                      downmand);

    g_signal_connect (project, "transfer_new",
                      G_CALLBACK (downmand_transfer_new),
                      downmand);

    g_signal_connect (project, "transfer_stop",
                      G_CALLBACK (downmand_transfer_stop),
                      downmand);

    g_signal_connect (project, "transfer_done",
                      G_CALLBACK (downmand_transfer_done),
                      downmand);

    g_signal_connect (project, "transfer_delete",
                      G_CALLBACK (downmand_transfer_delete),
                      downmand);

    g_signal_connect (project, "transfer_update",
                      G_CALLBACK (downmand_transfer_update),
                      downmand);
}

void
downmand_command_project_delete (Downmand *downmand, Project *project,
                                 GError **error)
{
    GString *buffer = downmand->buffer;
    GError *err = NULL;

    if (g_hash_table_size (downmand->projects) == 1)
    {
        g_set_error (error, DM_DOWNMAND_ERROR, 1,
            "You need at least one project");
        return;
    }

    if (project_size (project) > 0)
    {
        g_set_error (error, DM_DOWNMAND_ERROR, 1,
            "Remove all the transfers before deleting the project");
    }
    else
    {
        GDir *directory;
        const gchar *file;

        /* notice clients */
        downmand_clients_project_delete (downmand, project);

        g_hash_table_remove (downmand->projects,
                             GINT_TO_POINTER (project_id (project)));
        directory = g_dir_open (project_path_get (project), 0, &err);

        if (directory == NULL)
        {
            g_set_error (error, DM_DOWNMAND_ERROR, 1,
                         "Error deleting the project: %s",
                         err->message);
            return;
        }

        for (file = g_dir_read_name (directory);
             file != NULL;
             file = g_dir_read_name (directory))
        {
            g_string_printf (buffer, "%s/%s", project_path_get (project), file);
            unlink (buffer->str);
        }

        if (rmdir (project_path_get (project)) == -1)
        {
            g_set_error (error, DM_DOWNMAND_ERROR, 1,
                "Error while deleting the project %s", project_name(project));
        }
//FIXME: project_delete
//			project_delete(project);
//  unref the project and check if it is deleted
    }
}

/* Extracts the project and the transfer id from the command */
void
downmand_command_get_project (Downmand *downmand, gchar *command,
                              Project **project,
                              GError **error)
{
	guint project_id;

    if (sscanf(command, "%u", &project_id) > 0)
    {
        if (project_id > 0)
        {
            *project = g_hash_table_lookup (downmand->projects,
                                            GINT_TO_POINTER (project_id));

            if (*project == NULL)
            {
                g_set_error (error, DM_DOWNMAND_ERROR, 1,
                    "Project %i does not exist", project_id);
            }
        }
        else
        {
            g_set_error (error, DM_DOWNMAND_ERROR, 1,
                "Command does not contain a valid project id");
        }
    }
    else
    {
        g_set_error (error, DM_DOWNMAND_ERROR, 1,
            "Command does not contain a project id");
    }
}

/* Extracts the project and the guint setting from the command */
void
downmand_command_get_project_and_setting (Downmand *downmand, gchar *command,
                                          Project **project,
                                          guint *setting,
                                          GError **error)
{
	guint project_id;

    if (sscanf(command, "%u %u", &project_id, setting) > 1)
    {
        if (project_id > 0)
        {
            *project = g_hash_table_lookup (downmand->projects,
                                            GINT_TO_POINTER (project_id));

            if (*project == NULL)
            {
                g_set_error (error, DM_DOWNMAND_ERROR, 1,
                    "Project %i does not exist", project_id);
            }
        }
        else
        {
            g_set_error (error, DM_DOWNMAND_ERROR, 1,
                "Command does not contain a project id");
        }
    }
    else
    {
        g_set_error (error, DM_DOWNMAND_ERROR, 1,
            "Command does not contain a project id or setting");
    }
}

/* Extracts the guint setting from the command */
void
downmand_command_get_setting (gchar *command, guint *setting,
                              GError **error)
{
    if (sscanf(command, "%u", setting) < 1)
    {
        g_set_error (error, DM_DOWNMAND_ERROR, 1,
            "Command does not contain a setting");
    }

	return;
}


/* Servers ********************************************************************/

gint
downmand_servers_read (Downmand *downmand, GError **error)
{
    xmlDocPtr doc;
    xmlNodePtr node;
    Server *server;
    GString *buffer = downmand->buffer;
    xmlChar *value;

    g_string_printf (buffer, "%s/.downman/servers", g_get_home_dir ());
    doc = xmlParseFile (buffer->str);

    if (doc)
    {
        if (!doc->children || !doc->children->name
            || strcasecmp(doc->children->name, "TransferManagerServers") != 0)
        {
            xmlFreeDoc (doc);
            g_set_error (error, DM_DOWNMAND_ERROR, 1,
                         "Servers file is not valid: %s", buffer->str);
            return 1;
        }
        else
        {
            for (node = doc->children->children; node != NULL; node = node->next)
            {
                if (!node->name || (strcasecmp (node->name, "Server") != 0))
                    continue;

                server = g_new0 (Server, 1);

                value = xmlGetProp (node, "max_conn");
                server->max_connections = atol (value);
                xmlFree (value);

                value = xmlGetProp (node, "server_url");
                server->server_url = g_strdup (value);
                xmlFree (value);

                servers_server_new (downmand->servers, server);
            }
        }
    }

    xmlFreeDoc (doc);

    /* setup signals */
    g_signal_connect (downmand->servers, "servers_new",
                      G_CALLBACK (downmand_servers_new),
                      downmand);
    g_signal_connect (downmand->servers, "servers_changed",
                      G_CALLBACK (downmand_servers_changed),
                      downmand);

    return 0;
}

void
downmand_servers_save_one (gpointer key, gpointer value, gpointer user_data)
{
    xmlDocPtr   doc    = (xmlDocPtr) user_data;
    gchar      *name   = (gchar   *) key;
    Server     *server = (Server  *) value;
    xmlNodePtr  newxml;
    GString    *buffer;

    /* only save servers that have a connection limit */
    if (server->max_connections == -1)
        return;

    buffer = g_string_new (NULL);
    newxml = xmlNewChild (doc->children, NULL, "Server", NULL);
    g_string_printf (buffer, "%i", server->max_connections);
    xmlSetProp (newxml, "max_conn", buffer->str);
    xmlSetProp (newxml, "server_url", name);
    g_string_free (buffer, TRUE);
}

gboolean
downmand_servers_save (gpointer data)
{
    Downmand *downmand = DOWNMAND (data);
    xmlDocPtr doc;
    GString *buffer = downmand->buffer;

    doc = xmlNewDoc ("1.0");
    doc->children = xmlNewDocNode (doc, NULL, "TransferManagerServers", NULL);

    g_hash_table_foreach (downmand->servers->hash,
                          downmand_servers_save_one,
                          doc);

//FIXME: save to a different file and test for integrity
    g_string_printf (buffer, "%s/.downman/servers", g_get_home_dir ());

    if (xmlSaveFile (buffer->str, doc) == -1)
    {
// FIXME: better error handling
//		return 1;
    }
    xmlFreeDoc(doc);

    return TRUE;
}


/* Projects *******************************************************************/

gint
downmand_projects_read (Downmand *downmand, GError **error)
{
	xmlDocPtr doc;
	xmlNodePtr node;
	Project *project;
	GString *buffer = downmand->buffer;
	int projects = 0;
    xmlChar *value;
    GError *err = NULL;

	g_string_printf(buffer, "%s/.downman/projects_list", g_get_home_dir ());
	doc = xmlParseFile(buffer->str);

	if (doc)
	{
		if (!doc->children || !doc->children->name
			|| strcasecmp(doc->children->name, "TransferManagerProjects") != 0)
		{
            xmlFreeDoc(doc);
            g_set_error (error, DM_DOWNMAND_ERROR, 1,
                         "Projects file is not valid: %s", buffer->str);
            return 1;
		}
		else
		{
			for (node = doc->children->children; node != NULL; node = node->next)
			{
				if (!node->name || (strcasecmp(node->name, "Project") != 0))
					continue;

                project = project_new (downmand_project_id (downmand),
                                       &err);

                if (err != NULL)
                {
                    g_propagate_error (error, err);
                    continue;
                }

                value = xmlGetProp (node, "Name");
                project_name_set (project, value);
                xmlFree (value);

                value = xmlGetProp (node, "Path");
                project_default_path_set (project, value);
                xmlFree (value);

                downmand_project_new (downmand, project, &err);

                if (err != NULL)
                {
                    g_propagate_error (error, err);
                    return 1;
                }
                projects++;
			}
		}
	}

    if (projects == 0)
    {
        /* add default project */
        project = project_new (downmand_project_id (downmand), &err);

        if (err != NULL)
        {
            g_propagate_error (error, err);
            return 1;
        }

        project_name_set (project, "Main");
        g_string_printf(buffer, "%s/transfers/Main/", g_get_home_dir ());
        project_default_path_set (project, buffer->str);

        downmand_project_new (downmand, project, &err);

        if (err != NULL)
        {
            g_propagate_error (error, err);
            return 1;
        }
    }

    xmlFreeDoc(doc);

    return 0;
}

void
downmand_projects_save_one (gpointer key, gpointer value, gpointer user_data)
{
    Project *project = PROJECT (value);
    xmlDocPtr doc = (xmlDocPtr) user_data;
	xmlNodePtr newxml;

    /* save project preferences */
    project_preferences_save (project);
    project_session_save (project);

    /* save project name and path*/
    newxml = xmlNewChild (doc->children, NULL, "Project", NULL);
    xmlSetProp (newxml, "Name", project_name_get (project));
    xmlSetProp (newxml, "Path", project_default_path_get (project));
}

gboolean
downmand_projects_save (gpointer data)
{
	Downmand *downmand = (Downmand *) data;
	xmlDocPtr doc;
	GString *buffer = g_string_new(NULL);

	doc = xmlNewDoc("1.0");
	doc->children = xmlNewDocNode(doc, NULL, "TransferManagerProjects", NULL);

    g_hash_table_foreach (downmand->projects,
                          downmand_projects_save_one,
                          doc);

//FIXME: save to a different file and test for integrity
    g_string_printf (buffer, "%s/.downman/projects_list", g_get_home_dir ());

    if (xmlSaveFile (buffer->str, doc) == -1)
    {
// FIXME: better error handling
//		return 1;
    }
    xmlFreeDoc(doc);
    g_string_free (buffer, TRUE);

    return TRUE;
}

void
downmand_projects_scheduler_one (gpointer key, gpointer value,
                                 gpointer user_data)
{
    project_scheduler (PROJECT (value));
}

/* Simply call project_scheduler on every Project */
gboolean
downmand_projects_scheduler (gpointer data)
{
	Downmand *downmand = (Downmand *) data;

    g_hash_table_foreach (downmand->projects,
                          downmand_projects_scheduler_one,
                          NULL);

	return TRUE;
}


/* Clients ********************************************************************/

gboolean
downmand_new_client (GIOChannel *source, GIOCondition condition, gpointer data)
{
    Downmand *downmand = DOWNMAND (data);
    struct sockaddr_un addr;
    int size = sizeof (addr);
    int fd;

    if ((condition & G_IO_HUP) || (condition & G_IO_ERR) ||
        (condition & G_IO_PRI) || (condition & G_IO_NVAL) ||
        (condition & G_IO_OUT))
    {
        g_debug ("Bad condition on new connection\n");
    }
    else if (condition & G_IO_IN)
    {
        fd = accept (downmand->local_socket, (struct sockaddr *)&addr, &size);

        if (fd < 0)
        {
            g_debug("Can't accept client connection, error was: %s\n", g_strerror(errno));
        }
        else
        {
            Client *client = client_new (fd);

            g_return_val_if_fail (client != NULL, TRUE);

            g_signal_connect (client, "command_received",
                              G_CALLBACK (downmand_command_handle),
                              downmand);

            g_signal_connect (client, "disconnect",
                              G_CALLBACK (downmand_client_delete),
                              downmand);

            client->id = ++downmand->clients_count;
            g_hash_table_insert (downmand->clients,
                                 GINT_TO_POINTER (client->id),
                                 client);
        }

        return TRUE;
    }

    /* We shouldn't get here */
    g_return_val_if_reached(TRUE);
}

void
downmand_client_delete (GObject *caller, gpointer data)
{
    Downmand *downmand = DOWNMAND (data);
    Client *client = CLIENT (caller);

    g_hash_table_remove (downmand->clients, GINT_TO_POINTER (client->id));
    g_object_unref (G_OBJECT (client));
}

void
downmand_events_clients_one_send (gpointer data, gpointer user_data)
{
    Client *client = CLIENT (user_data);
    Transfer *transfer = TRANSFER (data);
    GString *buffer = client->buffer;
    guint project_id;

//FIXME: remove hack
    project_id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (client),
                                                     "project_id"));

    g_string_printf (buffer, "%i %u %i %i %i %Li %Li\n",
                     DM_COMMAND_BATCH_PROGRESS,
                     project_id,
                     transfer_id_get (transfer),
                     transfer_status_get (transfer),
                     transfer_speed (transfer),
                     transfer_bytes_done_get (transfer),
                     transfer_bytes_total_get (transfer));

    client_send (client, buffer->str);
}

void
downmand_events_clients_one (gpointer key, gpointer value, gpointer user_data)
{
    Project *project = PROJECT (user_data);
    Client *client = CLIENT (value);
    GString *buffer = client->buffer;

//FIXME: this is a quick hack, switch to global transfer id
    g_object_set_data (G_OBJECT (client), "project_id",
                       GINT_TO_POINTER (project_id (project)));

    /* test for batch status */
    if (client_batch_status (client))
    {
        g_string_printf (buffer, "%i %u %i %i %Lu %Lu\n",
                        DM_COMMAND_BATCH_STATUS,
                        project_id (project),
                        project->active_transfers,
                        project_queued_transfers (project),
                        project->bytes_done,
                        project->bytes_total);
        client_send (client, buffer->str);
    }

    /* test for progress */
    if (client_batch_progress (client))
    {
        g_list_foreach (project_batch_active (project),
                        downmand_events_clients_one_send,
                        client);
    }
}

void
downmand_events_projects_one_totals (gpointer data, gpointer user_data)
{
    Project *project = PROJECT (user_data);
    Transfer *transfer = TRANSFER (data);

    project->active_transfers++;
    project->bytes_done += transfer_bytes_done_get (transfer);
    project->bytes_total += transfer_bytes_total_get (transfer);
}

void
downmand_events_projects_one (gpointer key, gpointer value, gpointer user_data)
{
    Downmand *downmand = DOWNMAND (user_data);
    Project *project = PROJECT (value);

    project->active_transfers = 0;
    project->bytes_total = 0;
    project->bytes_total = 0;

    /* Update transfers statistics */
    g_list_foreach (project_batch_active (project),
                    downmand_events_projects_one_totals,
                    project);

    /* progress all clients */
    g_hash_table_foreach (downmand->clients,
                          downmand_events_clients_one,
                          project);
}

gboolean
downmand_clients_test_events (gpointer data)
{
    Downmand *downmand = DOWNMAND (data);

    g_hash_table_foreach (downmand->projects,
                          downmand_events_projects_one,
                          downmand);

    return TRUE;
}

void
downmand_clients_batch_send_one_project (gpointer key, gpointer value,
                                         gpointer user_data)
{
    Client *client = CLIENT (user_data);
    Transfer *transfer = TRANSFER (value);
	GString *buffer = client->buffer;
    guint project_id;

//FIXME: remove hack
    project_id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (client),
                                                     "project_id"));

    g_string_printf (buffer, "%i %u %i %s %i %s %Li %Li %u\n",
                     DM_COMMAND_ITEM_NEW,
                     project_id,
                     transfer_id_get (transfer),
                     transfer_target_uri_string (transfer),
                     transfer_status_get (transfer),
                     transfer_source_uri_string (transfer),
                     transfer_bytes_done_get (transfer),
                     transfer_bytes_total_get (transfer),
                     transfer_speedlimit_get (transfer));
    client_send (client, buffer->str);
}

void
downmand_clients_batch_send_one (gpointer key, gpointer value,
                                 gpointer user_data)
{
    Client *client = CLIENT (user_data);
    Project *project = PROJECT (value);
	GString *buffer = client->buffer;

    g_string_printf (buffer, "%i %i %s\n", DM_COMMAND_PROJECT_NEW,
                     project_id (project),
                     project_name (project));
    client_send (client, buffer->str);

    g_string_printf (buffer, "%i %u %s\n", DM_COMMAND_PROJECT_PATH,
                     project_id (project),
                     project_default_path_get (project));
    client_send (client, buffer->str);

    g_string_printf (buffer, "%i %u %u\n", DM_COMMAND_PROJECT_AUTO,
                     project_id (project),
                     project_auto (project));
    client_send (client, buffer->str);

    g_string_printf (buffer, "%i %u %u\n", DM_COMMAND_PROJECT_MAXTRANSFERS,
                     project_id (project),
                     project->max_transfers);
    client_send (client, buffer->str);

//FIXME: this is a quick hack, switch to global transfer id
    g_object_set_data (G_OBJECT (client), "project_id",
                       GINT_TO_POINTER (project_id (project)));

    /* send every transfer on the project */
    g_hash_table_foreach (project->hash,
                          downmand_clients_batch_send_one_project,
                          client);
}

void
downmand_clients_batch_send (Downmand *downmand, Client *client)
{
    g_hash_table_foreach (downmand->projects,
                          downmand_clients_batch_send_one,
                          client);
}

void
downmand_clients_server_send (gpointer key, gpointer value, gpointer user_data)
{
    Client  *client = CLIENT (user_data);
    gchar   *name   = (gchar *) key;
    Server  *server = (Server *) value;
    GString *buffer = g_string_new (NULL);

    g_string_printf (buffer, "%i %s %i %u\n", DM_COMMAND_SERVER_NEW,
                     name,
                     server->max_connections,
                     server->current_connections);

    client_send (client, buffer->str);
}

void
downmand_clients_servers_send (Downmand *downmand, Client *client)
{
    g_hash_table_foreach (downmand->servers->hash,
                          downmand_clients_server_send,
                          client);
}


/* Signal handlers ************************************************************/

/* This function is called when a project changes the autotransfer setting.
   It builds a command and sends it to all clients configured to get
   project events.
 */
void
downmand_project_autotransfer_send (Project *project, gpointer data)
{
    Downmand *downmand = DOWNMAND (data);
	GString *buffer = g_string_new (NULL);

    g_string_printf (buffer, "%i %i %u\n", DM_COMMAND_PROJECT_AUTO,
                     project_id(project),
                     project_auto(project));

    downmand_project_send_to_clients (downmand, buffer->str);

    g_string_free (buffer, TRUE);
}

/* This function is called when a project changes the maxtransfers setting.
   It builds a command and sends it to all clients configured to get
   project events.
 */
void
downmand_project_maxtransfers_send (Project *project, gpointer data)
{
    Downmand *downmand = DOWNMAND (data);
	GString *buffer = g_string_new (NULL);

    g_string_printf (buffer, "%i %i %u\n", DM_COMMAND_PROJECT_MAXTRANSFERS,
                     project_id(project),
                     project_maxtransfers_get (project));

    downmand_project_send_to_clients (downmand, buffer->str);

    g_string_free (buffer, TRUE);
}

/* This function is called when a transfer is added to a project.
   It builds a command and sends it to all clients configured to get events.
 */
void downmand_transfer_new(Project *project, Transfer *transfer,
                           gpointer data)
{
    Downmand *downmand = DOWNMAND(data);
	GString *buffer = g_string_new(NULL);

    g_string_printf(buffer, "%i %u %i %s %i %s %Li %Li %u\n",
                    DM_COMMAND_ITEM_NEW,
                    project_id (project),
                    transfer_id_get (transfer),
                    transfer_target_uri_string (transfer),
                    transfer_status_get (transfer),
                    transfer_source_uri_string (transfer),
                    transfer_bytes_done_get (transfer),
                    transfer_bytes_total_get (transfer),
                    transfer_speedlimit_get (transfer));

    downmand_send_to_clients(downmand, buffer->str);

    g_string_free(buffer, TRUE);
}

/* This function is called when a transfer is stopped in a project.
   It builds a command and sends it to all clients configured to get events.
 */
void downmand_transfer_stop(Project *project, Transfer *transfer,
                            gpointer data)
{
    Downmand *downmand = DOWNMAND(data);
	GString *buffer = g_string_new(NULL);

    g_string_printf(buffer, "%i %u %i %i %Li %Li\n", DM_COMMAND_ITEM_STOP,
                    project_id (project),
                    transfer_id_get (transfer),
                    transfer_status_get (transfer),
                    transfer_bytes_done_get (transfer),
                    transfer_bytes_total_get (transfer));

    downmand_send_to_clients(downmand, buffer->str);

    g_string_free(buffer, TRUE);
}

/* This function is called when a transfer is finished.
   It builds a command and sends it to all clients configured to get events.
 */
void downmand_transfer_done(Project *project, Transfer *transfer,
                            gpointer data)
{
    Downmand *downmand = DOWNMAND(data);
	GString *buffer = g_string_new(NULL);

    g_string_printf(buffer, "%i %u %i %i %Li %Li\n", DM_COMMAND_ITEM_DONE,
                    project_id (project),
                    transfer_id_get (transfer),
                    STATUS_DONE,
                    transfer_bytes_done_get (transfer),
                    transfer_bytes_total_get (transfer));

    downmand_send_to_clients(downmand, buffer->str);

    g_string_free(buffer, TRUE);
}

/* This function is called when a transfer is deleted from a project.
   It builds a command and sends it to all clients configured to get events.
 */
void downmand_transfer_delete(Project *project, Transfer *transfer,
                              gpointer data)
{
    Downmand *downmand = DOWNMAND(data);
	GString *buffer = g_string_new(NULL);

    g_string_printf(buffer, "%i %u %i\n", DM_COMMAND_ITEM_DELETE,
                    project_id (project),
                    transfer_id_get (transfer));

    downmand_send_to_clients(downmand, buffer->str);

    g_string_free(buffer, TRUE);
}

/* This function is called when a transfer changes some of its configuration
   It builds a command and sends it to all clients configured to get events.
 */
void
downmand_transfer_update (Project *project, Transfer *transfer, gint setting,
                          gpointer data)
{
    Downmand *downmand = DOWNMAND(data);
    GString *buffer = g_string_new(NULL);

    switch (setting)
    {
        case DM_COMMAND_ITEM_SPEED:
            g_string_printf (buffer, "%i %u %i %i\n",
                             DM_COMMAND_ITEM_SPEED,
                             project_id (project),
                             transfer_id_get (transfer),
                             transfer_speedlimit_get (transfer));
            break;

        default:
            g_print ("update to transfer not handled!\n");
            break;
    }

    downmand_send_to_clients (downmand, buffer->str);

    g_string_free (buffer, TRUE);
}

void
downmand_servers_new (Servers *servers, Server *server, gpointer data)
{
    Downmand *downmand = DOWNMAND (data);
    GString *buffer = g_string_new (NULL);

    g_string_printf (buffer, "%i %s %i %u\n",
                     DM_COMMAND_SERVER_NEW,
                     server->server_url,
                     server->max_connections,
                     server->current_connections);

//FIXME: send only to clients that want SERVER events
    downmand_send_to_clients (downmand, buffer->str);

    g_string_free (buffer, TRUE);
}

void
downmand_servers_changed (Servers *servers, Server *server, gint setting,
                          gpointer data)
{
    Downmand *downmand = DOWNMAND (data);
    GString *buffer = g_string_new (NULL);

    switch (setting)
    {
        case DM_COMMAND_SERVER_MAX_CONNECTIONS:
            g_string_printf (buffer, "%i %s %i\n",
                             DM_COMMAND_SERVER_MAX_CONNECTIONS,
                             server->server_url,
                             server->max_connections);
            break;

        case DM_COMMAND_SERVER_CURRENT_CONNECTIONS:
            g_string_printf (buffer, "%i %s %u\n",
                             DM_COMMAND_SERVER_CURRENT_CONNECTIONS,
                             server->server_url,
                             server->current_connections);
            break;

        default:
            /* incomplete implementation */
            g_printerr ("update to servers not handled!\n");
            break;
    }

//FIXME: send only to clients that want SERVER events
    downmand_send_to_clients (downmand, buffer->str);

    g_string_free (buffer, TRUE);
}

void
downmand_send_to_clients_one (gpointer key, gpointer value, gpointer user_data)
{
    Client *client = CLIENT (value);
    gchar *command = (gchar *) user_data;

    /* test for events */
    if (client_events (client))
        client_send (client, command);
}

/* Send the command to all clients configured to get events.
 */
void
downmand_send_to_clients (Downmand *downmand, gchar *command)
{
    g_hash_table_foreach (downmand->clients,
                          downmand_send_to_clients_one,
                          command);
}

void
downmand_project_send_to_clients_one (gpointer key, gpointer value,
                                      gpointer user_data)
{
    Client *client = CLIENT (value);
    gchar *command = (gchar *) user_data;

    /* test for events */
    if (client_events (client) || client_events_projects (client))
        client_send (client, command);
}

/* Send the command to all clients configured to get project events.
 */
void
downmand_project_send_to_clients (Downmand *downmand, gchar *command)
{
    g_hash_table_foreach (downmand->clients,
                          downmand_project_send_to_clients_one,
                          command);
}

void
downmand_clients_projects_send_one (gpointer key, gpointer value,
                                    gpointer user_data)
{
    Client *client = CLIENT (user_data);
    Project *project = PROJECT (value);
	GString *buffer = client->buffer;

    g_string_printf (buffer, "%i %i %s\n", DM_COMMAND_PROJECT_NEW,
                     project_id (project),
                     project_name (project));

    client_send (client, buffer->str);
}

void
downmand_clients_projects_send (Downmand *downmand, Client *client)
{
    g_hash_table_foreach (downmand->projects,
                          downmand_clients_projects_send_one,
                          client);
}

void
downmand_clients_project_new (Downmand *downmand, Project *project)
{
    GString *buffer = downmand->buffer;

    g_string_printf (buffer, "%i %i %s\n", DM_COMMAND_PROJECT_NEW,
                     project_id(project),
                     project_name(project));

    downmand_project_send_to_clients (downmand, buffer->str);
}

void
downmand_clients_project_delete (Downmand *downmand, Project *project)
{
    GString *buffer = downmand->buffer;

    g_string_printf (buffer, "%i %i\n", DM_COMMAND_PROJECT_DELETE,
                     project_id(project));

    downmand_project_send_to_clients (downmand, buffer->str);
}

void
downmand_shutdown (Downmand *downmand)
{
    if (!g_source_remove (downmand->scheduler_timeout))
    {
        g_warning ("scheduler_timeout not destroyed\n");
    }

    if (!g_source_remove (downmand->clients_timeout))
    {
        g_warning ("clients_timeout not destroyed\n");
    }

    if (!g_source_remove (downmand->save_timeout))
    {
        g_warning ("save_timeout not destroyed\n");
    }

    downmand_shutdown_transfers_stop (downmand);

    downmand_projects_save (downmand);
}

void
downmand_shutdown_transfers_stop (Downmand *downmand)
{
//FIXME: better shutdown
/*
    Transfer *transfer;
    Node *element;

    element = list_begin(downmand->projects);

    while (element != 0)
    {
        Project *project = (Project *) list_node_data(element);
        List *batch_active = project_batch_active(project);
        Node *element2;

        element2 = list_begin(batch_active);
        while (element2 != 0)
        {
            transfer = (Transfer *)list_node_data(element2);

            transfer_stop (transfer, TRANSFER_STOP_SHUTDOWN);

            element2 = list_next(element2);
        }

        element = list_next(element);
    }

    return;
*/
}

void
sigterm_handler (int signal)
{
    g_print ("SIGTERM received...\n");
//    downmand_shutdown_set(downmand, 1);
}

void downmand_main_loop_conf(Downmand *downmand)
{
	downmand->main_loop = g_main_loop_new(NULL, FALSE);

	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR)
	{
        perror("can't ignore SIGPIPE");
        downmand_shutdown_set(downmand, 1);
        g_main_loop_quit(downmand->main_loop);
	}

    if (signal(SIGTERM, sigterm_handler) == SIG_ERR)
	{
        perror("can't ignore SIGTERM");
        downmand_shutdown_set(downmand, 1);
        g_main_loop_quit(downmand->main_loop);
	}

	g_io_add_watch(downmand->g_local_socket,
					G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
					downmand_new_client, downmand);

	/* Schedule every second */
    downmand->scheduler_timeout = g_timeout_add (1000,
                                                 downmand_projects_scheduler,
                                                 downmand);

	/* Send info to clients */
	downmand->clients_timeout = g_timeout_add (1000,
                                               downmand_clients_test_events,
                                               downmand);

	/* Save projects every minute */
    downmand->save_timeout = g_timeout_add (1000*20,
                                            downmand_projects_save,
                                            downmand);

    /* Save servers every 5 minutes */
    downmand->save_timeout = g_timeout_add (1000*60*5,
                                            downmand_servers_save,
                                            downmand);
}


int downmand_socket_local(Downmand *downmand)
{
	return downmand->local_socket;
}

GHashTable *
downmand_clients (Downmand *downmand)
{
    return downmand->clients;
}

GHashTable *
downmand_projects (Downmand *downmand)
{
    return downmand->projects;
}

gint
downmand_shutdown_get (Downmand *downmand)
{
    return downmand->shutdown;
}

void
downmand_shutdown_set (Downmand *downmand, gint i)
{
    downmand->shutdown = i;
}

GMainLoop * downmand_main_loop(Downmand *downmand)
{
	return downmand->main_loop;
}

GQuark dm_downmand_error_quark (void)
{
  static GQuark quark;
  if (!quark)
    quark = g_quark_from_static_string ("dm_downmand_error");
  return quark;
}
