#ifndef __TRANSFER_H__
#define __TRANSFER_H__

#include <glib-object.h>
#include <libxml/uri.h>
#include <libgnomevfs/gnome-vfs-uri.h>
#include <libgnomevfs/gnome-vfs-file-info.h>
#include <libgnomevfs/gnome-vfs-async-ops.h>

#include "linelog.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

typedef enum {
    TRANSFER_QUEUE_NONE = 0,
    TRANSFER_QUEUE_ACTIVE,
    TRANSFER_QUEUE_BATCH,
    TRANSFER_QUEUE_STOPPED,
    TRANSFER_QUEUE_DONE
} TransferQueue;

typedef enum {
    TRANSFER_EXIT_NONE = 0,
    TRANSFER_EXIT_SHUTDOWN,
    TRANSFER_EXIT_STOPPED,
    TRANSFER_EXIT_QUEUED,
    TRANSFER_EXIT_DELETE,
    TRANSFER_EXIT_DONE
} TransferExitCause;

typedef enum {
    TRANSFER_STOP_NONE,
    TRANSFER_STOP_SHUTDOWN,
    TRANSFER_STOP_DELETE,
    TRANSFER_STOP_QUEUE
} TransferStopAction;

#define TRANSFER(obj)           G_TYPE_CHECK_INSTANCE_CAST (obj, transfer_get_type (), Transfer)
#define TRANSFER_CLASS(klass)   G_TYPE_CHECK_CLASS_CAST (klass, transfer_get_type (), TransferClass)
#define IS_TRANSFER(obj)        G_TYPE_CHECK_INSTANCE_TYPE (obj, transfer_get_type ())


typedef struct _Transfer       Transfer;
typedef struct _TransferClass  TransferClass;

struct _Transfer
{
    GObject object;

    /*< private >*/

    guint id;
    guint order;

    GnomeVFSURI *source_uri;
    GnomeVFSFileInfo *source_info;

    GnomeVFSURI *target_uri;
    GnomeVFSFileInfo *target_info;

    GnomeVFSAsyncHandle *async_handle;

    GnomeVFSURI *target_path;

    gboolean default_path;
    gboolean target_created;

	GnomeVFSFileSize bytes_done;
	GnomeVFSFileSize bytes_total;

    guint speed;
	guint speeds[5];
    struct timeval oldtime;

    GList *threads;

    gboolean target_ready;
    TransferQueue queue;
//FIXME: convert enum to type
    enum dm_transfer_status status;
    TransferExitCause exit_cause;

    /* loggin facilities */
    Linelog *linelog;

    /* settings */
    guint recursive;
    /* the limit set on the transfer */
    guint speedlimit;
    /* the limit the transfer must obey (perhaps project mandated) */
    guint real_speedlimit;
};

struct _TransferClass
{
    GObjectClass parent_class;

    /* transfer_update_signal is emitted when some of the properties
     * of the transfer change (i.e. speedlimit)
     */
    void (* transfer_update_signal)     (gint setting);
    void (* transfer_exit_signal)       (void);
};

GType           transfer_get_type       (void);
Transfer*       transfer_new            ();

void    transfer_start      (Transfer *transfer, GError **error);
void    transfer_stop       (Transfer *transfer,
                             TransferStopAction action);
void    transfer_delete     (Transfer *transfer);

gboolean    transfer_source_uri_set (Transfer *transfer, gchar *url);
gboolean    transfer_target_uri_set (Transfer *transfer, gchar *url);
gboolean    transfer_path_set       (Transfer *transfer, gchar *path);

gint    transfer_status_get (Transfer *transfer);
void    transfer_status_set (Transfer *transfer,
                             enum dm_transfer_status status);

guint   transfer_recursive_get          (Transfer *transfer);
void    transfer_recursive_set          (Transfer *transfer, guint recursive);
void    transfer_default_path_set       (Transfer *transfer, gboolean setting);
void    transfer_speedlimit_set         (Transfer *transfer, guint speed);
guint   transfer_speedlimit_get         (Transfer *transfer);
void    transfer_real_speedlimit_set    (Transfer *transfer, guint speed);
guint   transfer_real_speedlimit_get    (Transfer *transfer);

TransferExitCause   transfer_exit_cause (Transfer *transfer);

guint   transfer_id_get     (Transfer *transfer);
void    transfer_id_set     (Transfer *transfer, guint id);

void    transfer_statistics (Transfer *transfer);

gchar*           transfer_source_uri_string  (Transfer *transfer);
gchar*           transfer_target_uri_string  (Transfer *transfer);
gchar*           transfer_targetpath_string  (Transfer *transfer);

GnomeVFSFileSize transfer_bytes_done_get     (Transfer *transfer);
void             transfer_bytes_done_set     (Transfer *transfer,
                                              GnomeVFSFileSize bytes_done);
GnomeVFSFileSize transfer_bytes_total_get    (Transfer *transfer);
void             transfer_bytes_total_set    (Transfer *transfer,
                                              GnomeVFSFileSize bytes_total);
guint            transfer_speed              (Transfer *transfer);


TransferQueue   transfer_queue_get  (Transfer *transfer);
void            transfer_queue_set  (Transfer *transfer,
                                     TransferQueue queue);

void            transfer_logdir_set (Transfer *transfer, gchar *path);
void            transfer_log_add    (Transfer *transfer, const gchar *format,
                                     ...);

void        transfer_target_created_set (Transfer *transfer, gboolean created);
gboolean    transfer_target_created_get (Transfer *transfer);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __TRANSFER_H__ */
